@@ -1,16 +1,16 @@
 /***************************************************************************
- *                                  _   _ ____  _     
- *  Project                     ___| | | |  _ \| |    
- *                             / __| | | | |_) | |    
- *                            | (__| |_| |  _ <| |___ 
+ *                                  _   _ ____  _
+ *  Project                     ___| | | |  _ \| |
+ *                             / __| | | | |_) | |
+ *                            | (__| |_| |  _ <| |___
  *                             \___|\___/|_| \_\_____|
  *
- * Copyright (C) 1998 - 2004, Daniel Stenberg, <daniel@haxx.se>, et al.
+ * Copyright (C) 1998 - 2005, Daniel Stenberg, <daniel@haxx.se>, et al.
  *
  * This software is licensed as described in the file COPYING, which
  * you should have received as part of this distribution. The terms
  * are also available at http://curl.haxx.se/docs/copyright.html.
- * 
+ *
  * You may opt to use, copy, modify, merge, publish, distribute and/or sell
  * copies of the Software, and permit persons to whom the Software is
  * furnished to do so, under the terms of the COPYING file.
@@ -18,7 +18,7 @@
  * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
  * KIND, either express or implied.
  *
- * $Id: connect.c,v 1.101 2004/05/13 10:40:17 bagder Exp $
+ * $Id: connect.c,v 1.139 2005/10/25 13:15:11 bagder Exp $
  ***************************************************************************/
 
 #include "setup.h"
@@ -67,7 +67,7 @@
 #undef in_addr_t
 #define in_addr_t unsigned long
 #endif
-#ifdef	VMS
+#ifdef VMS
 #include <in.h>
 #include <inet.h>
 #endif
@@ -97,12 +97,20 @@
 #include "strerror.h"
 #include "connect.h"
 #include "memory.h"
+#include "select.h"
+#include "url.h" /* for Curl_safefree() */
 
 /* The last #include file should be: */
 #include "memdebug.h"
 
 static bool verifyconnect(curl_socket_t sockfd, int *error);
 
+static curl_socket_t
+singleipconnect(struct connectdata *conn,
+                Curl_addrinfo *ai, /* start connecting to this */
+                long timeout_ms,
+                bool *connected);
+
 /*
  * Curl_ourerrno() returns the errno (or equivalent) on this platform to
  * hide platform specific for the function that calls this.
@@ -125,6 +133,7 @@ int Curl_nonblock(curl_socket_t sockfd,    /* operate on this */
                   int nonblock   /* TRUE or FALSE */)
 {
 #undef SETBLOCK
+#define SETBLOCK 0
 #ifdef HAVE_O_NONBLOCK
   /* most recent unix versions */
   int flags;
@@ -134,45 +143,52 @@ int Curl_nonblock(curl_socket_t sockfd,    /* operate on this */
     return fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);
   else
     return fcntl(sockfd, F_SETFL, flags & (~O_NONBLOCK));
+#undef SETBLOCK
 #define SETBLOCK 1
 #endif
 
-#ifdef HAVE_FIONBIO
+#if defined(HAVE_FIONBIO) && (SETBLOCK == 0)
   /* older unix versions */
   int flags;
 
   flags = nonblock;
   return ioctl(sockfd, FIONBIO, &flags);
+#undef SETBLOCK
 #define SETBLOCK 2
 #endif
 
-#ifdef HAVE_IOCTLSOCKET
+#if defined(HAVE_IOCTLSOCKET) && (SETBLOCK == 0)
   /* Windows? */
   unsigned long flags;
   flags = nonblock;
+
   return ioctlsocket(sockfd, FIONBIO, &flags);
+#undef SETBLOCK
 #define SETBLOCK 3
 #endif
 
-#ifdef HAVE_IOCTLSOCKET_CASE
+#if defined(HAVE_IOCTLSOCKET_CASE) && (SETBLOCK == 0)
   /* presumably for Amiga */
   return IoctlSocket(sockfd, FIONBIO, (long)nonblock);
+#undef SETBLOCK
 #define SETBLOCK 4
 #endif
 
-#ifdef HAVE_SO_NONBLOCK
+#if defined(HAVE_SO_NONBLOCK) && (SETBLOCK == 0)
   /* BeOS */
   long b = nonblock ? 1 : 0;
   return setsockopt(sockfd, SOL_SOCKET, SO_NONBLOCK, &b, sizeof(b));
+#undef SETBLOCK
 #define SETBLOCK 5
 #endif
 
 #ifdef HAVE_DISABLED_NONBLOCKING
   return 0; /* returns success */
+#undef SETBLOCK
 #define SETBLOCK 6
 #endif
 
-#ifndef SETBLOCK
+#if (SETBLOCK == 0)
 #error "no non-blocking method was found/used/set"
 #endif
 }
@@ -195,9 +211,6 @@ static
 int waitconnect(curl_socket_t sockfd, /* socket */
                 long timeout_msec)
 {
-  fd_set fd;
-  fd_set errfd;
-  struct timeval interval;
   int rc;
 #ifdef mpeix
   /* Call this function once now, and ignore the results. We do this to
@@ -207,27 +220,16 @@ int waitconnect(curl_socket_t sockfd, /* socket */
 #endif
 
   /* now select() until we get connect or timeout */
-  FD_ZERO(&fd);
-  FD_SET(sockfd, &fd);
-
-  FD_ZERO(&errfd);
-  FD_SET(sockfd, &errfd);
-
-  interval.tv_sec = timeout_msec/1000;
-  timeout_msec -= interval.tv_sec*1000;
-
-  interval.tv_usec = timeout_msec*1000;
-
-  rc = select(sockfd+1, NULL, &fd, &errfd, &interval);
+  rc = Curl_select(CURL_SOCKET_BAD, sockfd, (int)timeout_msec);
   if(-1 == rc)
     /* error, no connect here, try next */
     return WAITCONN_SELECT_ERROR;
-  
+
   else if(0 == rc)
     /* timeout, no connect today */
     return WAITCONN_TIMEOUT;
 
-  if(FD_ISSET(sockfd, &errfd))
+  if(rc & CSELECT_ERR)
     /* error condition caught */
     return WAITCONN_FDSET_ERROR;
 
@@ -255,7 +257,7 @@ static CURLcode bindlocal(struct connectdata *conn,
 
     /* First check if the given name is an IP address */
     in=inet_addr(data->set.device);
-      
+
     if((in == CURL_INADDR_NONE) &&
        Curl_if2ip(data->set.device, myhost, sizeof(myhost))) {
       /*
@@ -330,13 +332,16 @@ static CURLcode bindlocal(struct connectdata *conn,
         Curl_resolv_unlock(data, h);
         /* we don't need it anymore after this function has returned */
 
-#ifdef ENABLE_IPV6
-        if( bind(sockfd, addr->ai_addr, addr->ai_addrlen) >= 0) {
+        if( bind(sockfd, addr->ai_addr, (socklen_t)addr->ai_addrlen) >= 0) {
           /* we succeeded to bind */
+#ifdef ENABLE_IPV6
           struct sockaddr_in6 add;
+#else
+          struct sockaddr_in add;
+#endif
 
           bindworked = TRUE;
-	
+
           size = sizeof(add);
           if(getsockname(sockfd, (struct sockaddr *) &add,
                          (socklen_t *)&size)<0) {
@@ -344,51 +349,32 @@ static CURLcode bindlocal(struct connectdata *conn,
             return CURLE_HTTP_PORT_FAILED;
           }
         }
-#else
-        {
-          struct sockaddr_in sa;
 
-          memset((char *)&sa, 0, sizeof(sa));
-          memcpy((char *)&sa.sin_addr, addr->h_addr, addr->h_length);
-          sa.sin_family = AF_INET;
-          sa.sin_addr.s_addr = in;
-          sa.sin_port = 0; /* get any port */
-	
-          if( bind(sockfd, (struct sockaddr *)&sa, sizeof(sa)) >= 0) {
-            /* we succeeded to bind */
-            struct sockaddr_in add;
-	
-            bindworked = TRUE;
-            
-            size = sizeof(add);
-            if(getsockname(sockfd, (struct sockaddr *) &add,
-                           (socklen_t *)&size)<0) {
-              failf(data, "getsockname() failed");
-              return CURLE_HTTP_PORT_FAILED;
-            }
-          }
-        }
-#endif
         if(!bindworked) {
-          failf(data, "%s", Curl_strerror(conn, Curl_ourerrno()));
+          data->state.os_errno = Curl_ourerrno();
+          failf(data, "bind failure: %s",
+                Curl_strerror(conn, data->state.os_errno));
           return CURLE_HTTP_PORT_FAILED;
         }
-	
+
       } /* end of if  h */
       else {
-	failf(data,"could't find my own IP address (%s)", myhost);
-	return CURLE_HTTP_PORT_FAILED;
+        failf(data,"couldn't find my own IP address (%s)", myhost);
+        return CURLE_HTTP_PORT_FAILED;
       }
     } /* end of inet_addr */
 
     else {
-      failf(data, "could't find my own IP address (%s)", myhost);
+      failf(data, "couldn't find my own IP address (%s)", myhost);
       return CURLE_HTTP_PORT_FAILED;
     }
 
     return CURLE_OK;
 
   } /* end of device selection support */
+#else
+  (void)conn;
+  (void)sockfd;
 #endif /* end of HAVE_INET_NTOA */
 
   return CURLE_HTTP_PORT_FAILED;
@@ -420,13 +406,25 @@ static bool verifyconnect(curl_socket_t sockfd, int *error)
    *
    *    Someone got to verify this on Win-NT 4.0, 2000."
    */
+
+#ifdef _WIN32_WCE
+  Sleep(0);
+#else
   SleepEx(0, FALSE);
+#endif
+
 #endif
 
   if( -1 == getsockopt(sockfd, SOL_SOCKET, SO_ERROR,
                        (void *)&err, &errSize))
     err = Curl_ourerrno();
 
+#ifdef _WIN32_WCE
+  /* Always returns this error, bug in CE? */
+  if(WSAENOPROTOOPT==err)
+    err=0;
+#endif
+
   if ((0 == err) || (EISCONN == err))
     /* we are connected, awesome! */
     rc = TRUE;
@@ -443,43 +441,99 @@ static bool verifyconnect(curl_socket_t sockfd, int *error)
   return rc;
 }
 
+CURLcode Curl_store_ip_addr(struct connectdata *conn)
+{
+  char addrbuf[256];
+  Curl_printable_address(conn->ip_addr, addrbuf, sizeof(addrbuf));
+
+  /* save the string */
+  Curl_safefree(conn->ip_addr_str);
+  conn->ip_addr_str = strdup(addrbuf);
+  if(!conn->ip_addr_str)
+    return CURLE_OUT_OF_MEMORY; /* FAIL */
+
+#ifdef PF_INET6
+  if(conn->ip_addr->ai_family == PF_INET6)
+    conn->bits.ipv6 = TRUE;
+#endif
+
+  return CURLE_OK;
+}
+
+/* Used within the multi interface. Try next IP address, return TRUE if no
+   more address exists */
+static bool trynextip(struct connectdata *conn,
+                      int sockindex,
+                      bool *connected)
+{
+  curl_socket_t sockfd;
+  Curl_addrinfo *ai;
+
+  if(sockindex != FIRSTSOCKET)
+    return TRUE; /* no next */
+
+  /* first close the failed socket */
+  sclose(conn->sock[sockindex]);
+
+  /* try the next address */
+  ai = conn->ip_addr->ai_next;
+
+  while (ai) {
+    sockfd = singleipconnect(conn, ai, 0L, connected);
+    if(sockfd != CURL_SOCKET_BAD) {
+      /* store the new socket descriptor */
+      conn->sock[sockindex] = sockfd;
+      conn->ip_addr = ai;
+
+      Curl_store_ip_addr(conn);
+      return FALSE;
+    }
+    ai = ai->ai_next;
+  }
+  return TRUE;
+}
+
 /*
  * Curl_is_connected() is used from the multi interface to check if the
  * firstsocket has connected.
  */
 
 CURLcode Curl_is_connected(struct connectdata *conn,
-                           curl_socket_t sockfd,
+                           int sockindex,
                            bool *connected)
 {
   int rc;
   struct SessionHandle *data = conn->data;
+  CURLcode code = CURLE_OK;
+  curl_socket_t sockfd = conn->sock[sockindex];
+  long allow = DEFAULT_CONNECT_TIMEOUT;
+  long has_passed;
+
+  curlassert(sockindex >= FIRSTSOCKET && sockindex <= SECONDARYSOCKET);
 
   *connected = FALSE; /* a very negative world view is best */
 
-  if(data->set.timeout || data->set.connecttimeout) {
-    /* there is a timeout set */
+  /* Evaluate in milliseconds how much time that has passed */
+  has_passed = Curl_tvdiff(Curl_tvnow(), data->progress.t_startsingle);
 
-    /* Evaluate in milliseconds how much time that has passed */
-    long has_passed = Curl_tvdiff(Curl_tvnow(), data->progress.start);
-
-    /* subtract the most strict timeout of the ones */
-    if(data->set.timeout && data->set.connecttimeout) {
-      if (data->set.timeout < data->set.connecttimeout)
-        has_passed -= data->set.timeout*1000;
-      else 
-        has_passed -= data->set.connecttimeout*1000;
-    }
-    else if(data->set.timeout)
-      has_passed -= data->set.timeout*1000;
+  /* subtract the most strict timeout of the ones */
+  if(data->set.timeout && data->set.connecttimeout) {
+    if (data->set.timeout < data->set.connecttimeout)
+      allow = data->set.timeout*1000;
     else
-      has_passed -= data->set.connecttimeout*1000;
+      allow = data->set.connecttimeout*1000;
+  }
+  else if(data->set.timeout) {
+    allow = data->set.timeout*1000;
+  }
+  else if(data->set.connecttimeout) {
+    allow = data->set.connecttimeout*1000;
+  }
 
-    if(has_passed > 0 ) {
-      /* time-out, bail out, go home */
-      failf(data, "Connection time-out");
-      return CURLE_OPERATION_TIMEOUTED;
-    }
+  if(has_passed > allow ) {
+    /* time-out, bail out, go home */
+    failf(data, "Connection time-out after %ld ms", has_passed);
+    return CURLE_OPERATION_TIMEOUTED;
   }
   if(conn->bits.tcpconnect) {
     /* we are connected already! */
@@ -490,35 +544,50 @@ CURLcode Curl_is_connected(struct connectdata *conn,
   /* check for connect without timeout as we want to return immediately */
   rc = waitconnect(sockfd, 0);
 
-  if(0 == rc) {
-    if (verifyconnect(sockfd,NULL)) {
+  if(WAITCONN_CONNECTED == rc) {
+    int error;
+    if (verifyconnect(sockfd, &error)) {
       /* we are connected, awesome! */
       *connected = TRUE;
       return CURLE_OK;
     }
     /* nope, not connected for real */
-    failf(data, "Connection failed");
-    return CURLE_COULDNT_CONNECT;
+    data->state.os_errno = error;
+    infof(data, "Connection failed\n");
+    if(trynextip(conn, sockindex, connected)) {
+      code = CURLE_COULDNT_CONNECT;
+    }
   }
-  else if(1 != rc) {
-    int error = Curl_ourerrno();
-    failf(data, "Failed connect to %s:%d; %s",
-          conn->host.name, conn->port, Curl_strerror(conn,error));
-    return CURLE_COULDNT_CONNECT;
+  else if(WAITCONN_TIMEOUT != rc) {
+    int error = 0;
+
+    /* nope, not connected  */
+    if (WAITCONN_FDSET_ERROR == rc) {
+      verifyconnect(sockfd, &error);
+      data->state.os_errno = error;
+      infof(data, "%s\n",Curl_strerror(conn,error));
+    }
+    else
+      infof(data, "Connection failed\n");
+
+    if(trynextip(conn, sockindex, connected)) {
+      error = Curl_ourerrno();
+      data->state.os_errno = error;
+      failf(data, "Failed connect to %s:%d; %s",
+            conn->host.name, conn->port, Curl_strerror(conn,error));
+      code = CURLE_COULDNT_CONNECT;
+    }
   }
   /*
-   * If the connection phase is "done" here, we should attempt to connect
-   * to the "next address" in the Curl_hostaddr structure that we resolved
-   * before. But we don't have that struct around anymore and we can't just
-   * keep a pointer since the cache might in fact have gotten pruned by the
-   * time we want to read this... Alas, we don't do this yet.
+   * If the connection failed here, we should attempt to connect to the "next
+   * address" for the given host.
    */
 
-  return CURLE_OK;
+  return code;
 }
 
-static void Curl_setNoDelay(struct connectdata *conn,
-                            curl_socket_t sockfd)
+static void tcpnodelay(struct connectdata *conn,
+                       curl_socket_t sockfd)
 {
 #ifdef TCP_NODELAY
   struct SessionHandle *data= conn->data;
@@ -535,6 +604,125 @@ static void Curl_setNoDelay(struct connectdata *conn,
 #endif
 }
 
+#ifdef SO_NOSIGPIPE
+/* The preferred method on Mac OS X (10.2 and later) to prevent SIGPIPEs when
+   sending data to a dead peer (instead of relying on the 4th argument to send
+   being MSG_NOSIGNAL). Possibly also existing and in use on other BSD
+   systems? */
+static void nosigpipe(struct connectdata *conn,
+                      curl_socket_t sockfd)
+{
+  struct SessionHandle *data= conn->data;
+  int onoff = 1;
+  if(setsockopt(sockfd, SOL_SOCKET, SO_NOSIGPIPE, (void *)&onoff,
+                sizeof(onoff)) < 0)
+    infof(data, "Could not set SO_NOSIGPIPE: %s\n",
+          Curl_strerror(conn, Curl_ourerrno()));
+}
+#endif
+
+/* singleipconnect() connects to the given IP only, and it may return without
+   having connected if used from the multi interface. */
+static curl_socket_t
+singleipconnect(struct connectdata *conn,
+                Curl_addrinfo *ai,
+                long timeout_ms,
+                bool *connected)
+{
+  char addr_buf[128];
+  int rc;
+  int error;
+  bool conected;
+  struct SessionHandle *data = conn->data;
+  curl_socket_t sockfd;
+
+  sockfd = socket(ai->ai_family, conn->socktype, ai->ai_protocol);
+  if (sockfd == CURL_SOCKET_BAD)
+    return CURL_SOCKET_BAD;
+
+  *connected = FALSE; /* default is not connected */
+
+  Curl_printable_address(ai, addr_buf, sizeof(addr_buf));
+  infof(data, "  Trying %s... ", addr_buf);
+
+  if(data->set.tcp_nodelay)
+    tcpnodelay(conn, sockfd);
+
+#ifdef SO_NOSIGPIPE
+  nosigpipe(conn, sockfd);
+#endif
+  if(conn->data->set.device) {
+    /* user selected to bind the outgoing socket to a specified "device"
+       before doing connect */
+    CURLcode res = bindlocal(conn, sockfd);
+    if(res) {
+      sclose(sockfd); /* close socket and bail out */
+      return CURL_SOCKET_BAD;
+    }
+  }
+
+  /* set socket non-blocking */
+  Curl_nonblock(sockfd, TRUE);
+
+  /* Connect TCP sockets, bind UDP */
+  if(conn->socktype == SOCK_STREAM)
+    rc = connect(sockfd, ai->ai_addr, (socklen_t)ai->ai_addrlen);
+  else
+    rc = 0;
+
+  if(-1 == rc) {
+    error = Curl_ourerrno();
+
+    switch (error) {
+    case EINPROGRESS:
+    case EWOULDBLOCK:
+#if defined(EAGAIN) && EAGAIN != EWOULDBLOCK
+      /* On some platforms EAGAIN and EWOULDBLOCK are the
+       * same value, and on others they are different, hence
+       * the odd #if
+       */
+    case EAGAIN:
+#endif
+      rc = waitconnect(sockfd, timeout_ms);
+      break;
+    default:
+      /* unknown error, fallthrough and try another address! */
+      failf(data, "Failed to connect to %s: %s",
+            addr_buf, Curl_strerror(conn,error));
+      data->state.os_errno = error;
+      break;
+    }
+  }
+
+  /* The 'WAITCONN_TIMEOUT == rc' comes from the waitconnect(), and not from
+     connect(). We can be sure of this since connect() cannot return 1. */
+  if((WAITCONN_TIMEOUT == rc) &&
+     (data->state.used_interface == Curl_if_multi)) {
+    /* Timeout when running the multi interface */
+    return sockfd;
+  }
+
+  conected = verifyconnect(sockfd, &error);
+
+  if(!rc && conected) {
+    /* we are connected, awesome! */
+    *connected = TRUE; /* this is a true connect */
+    infof(data, "connected\n");
+    return sockfd;
+  }
+  else if(WAITCONN_TIMEOUT == rc)
+    infof(data, "Timeout\n");
+  else {
+    data->state.os_errno = error;
+    infof(data, "%s\n", Curl_strerror(conn, error));
+  }
+
+  /* connect failed or timed out */
+  sclose(sockfd);
+
+  return CURL_SOCKET_BAD;
+}
+
 /*
  * TCP connect to the given host with timeout, proxy or remote doesn't matter.
  * There might be more than one IP address to try out. Fill in the passed
@@ -543,36 +731,33 @@ static void Curl_setNoDelay(struct connectdata *conn,
 
 CURLcode Curl_connecthost(struct connectdata *conn,  /* context */
                           struct Curl_dns_entry *remotehost, /* use this one */
-                          int port,                  /* connect to this */
                           curl_socket_t *sockconn,   /* the connected socket */
-                          Curl_ipconnect **addr,     /* the one we used */
+                          Curl_addrinfo **addr,      /* the one we used */
                           bool *connected)           /* really connected? */
 {
   struct SessionHandle *data = conn->data;
-  int rc, error;
-  curl_socket_t sockfd= CURL_SOCKET_BAD;
-  int aliasindex=0;
-  char *hostname;
+  curl_socket_t sockfd = CURL_SOCKET_BAD;
+  int aliasindex;
+  int num_addr;
+  Curl_addrinfo *ai;
+  Curl_addrinfo *curr_addr;
 
   struct timeval after;
   struct timeval before = Curl_tvnow();
 
-#ifdef ENABLE_IPV6
-  struct addrinfo *ai;
-#endif  
-
   /*************************************************************
    * Figure out what maximum time we have left
    *************************************************************/
-  long timeout_ms=300000; /* milliseconds, default to five minutes */
+  long timeout_ms= DEFAULT_CONNECT_TIMEOUT;
+  long timeout_per_addr;
 
   *connected = FALSE; /* default to not connected */
 
   if(data->set.timeout || data->set.connecttimeout) {
-    double has_passed;
+    long has_passed;
 
     /* Evaluate in milliseconds how much time that has passed */
-    has_passed = Curl_tvdiff(Curl_tvnow(), data->progress.start);
+    has_passed = Curl_tvdiff(Curl_tvnow(), data->progress.t_startsingle);
 
 #ifndef min
 #define min(a, b)   ((a) < (b) ? (a) : (b))
@@ -582,7 +767,7 @@ CURLcode Curl_connecthost(struct connectdata *conn,  /* context */
     if(data->set.timeout && data->set.connecttimeout) {
       if (data->set.timeout < data->set.connecttimeout)
         timeout_ms = data->set.timeout*1000;
-      else 
+      else
         timeout_ms = data->set.connecttimeout*1000;
     }
     else if(data->set.timeout)
@@ -591,7 +776,7 @@ CURLcode Curl_connecthost(struct connectdata *conn,  /* context */
       timeout_ms = data->set.connecttimeout*1000;
 
     /* subtract the passed time */
-    timeout_ms -= (long)has_passed;
+    timeout_ms -= has_passed;
 
     if(timeout_ms < 0) {
       /* a precaution, no need to continue if time already is up */
@@ -600,124 +785,31 @@ CURLcode Curl_connecthost(struct connectdata *conn,  /* context */
     }
   }
 
-  hostname = data->change.proxy?conn->proxy.name:conn->host.name;
-  infof(data, "About to connect() to %s port %d\n",
-        hostname, port);
+  /* Max time for each address */
+  num_addr = Curl_num_addresses(remotehost->addr);
+  timeout_per_addr = timeout_ms / num_addr;
 
-#ifdef ENABLE_IPV6
-  /*
-   * Connecting with a getaddrinfo chain
-   */
-  for (ai = remotehost->addr; ai; ai = ai->ai_next, aliasindex++) {
-    sockfd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
-    if (sockfd == CURL_SOCKET_BAD)
-      continue;
+  ai = remotehost->addr;
 
-    else if(data->set.tcp_nodelay)
-      Curl_setNoDelay(conn, sockfd);
-#else
-  /*
-   * Connecting with old style IPv4-only support
+  /* Below is the loop that attempts to connect to all IP-addresses we
+   * know for the given host. One by one until one IP succeeds.
    */
 
-  /* This is the loop that attempts to connect to all IP-addresses we
-     know for the given host. One by one. */
-  for(rc=-1, aliasindex=0;
-      rc && (struct in_addr *)remotehost->addr->h_addr_list[aliasindex];
-      aliasindex++) {
-    struct sockaddr_in serv_addr;
+  if(data->state.used_interface == Curl_if_multi)
+    /* don't hang when doing multi */
+    timeout_per_addr = 0;
 
-    /* create an IPv4 TCP socket */
-    sockfd = socket(AF_INET, SOCK_STREAM, 0);
-    if(CURL_SOCKET_BAD == sockfd) {
-      failf(data, "couldn't create socket");
-      return CURLE_COULDNT_CONNECT; /* big time error */
-    }
+  /*
+   * Connecting with a Curl_addrinfo chain
+   */
+  for (curr_addr = ai, aliasindex=0; curr_addr;
+       curr_addr = curr_addr->ai_next, aliasindex++) {
 
-    else if(data->set.tcp_nodelay)
-      Curl_setNoDelay(conn, sockfd);
+    /* start connecting to the IP curr_addr points to */
+    sockfd = singleipconnect(conn, curr_addr, timeout_per_addr, connected);
 
-    /* nasty address work before connect can be made */
-    memset((char *) &serv_addr, '\0', sizeof(serv_addr));
-    memcpy((char *)&(serv_addr.sin_addr),
-           (struct in_addr *)remotehost->addr->h_addr_list[aliasindex],
-           sizeof(struct in_addr));
-    serv_addr.sin_family = remotehost->addr->h_addrtype;
-    serv_addr.sin_port = htons((unsigned short)port);
-#endif
-
-    if(conn->data->set.device) {
-      /* user selected to bind the outgoing socket to a specified "device"
-         before doing connect */
-      CURLcode res = bindlocal(conn, sockfd);
-      if(res)
-        return res;
-    }
-
-    /* set socket non-blocking */
-    Curl_nonblock(sockfd, TRUE);
-
-    /* do not use #ifdef within the function arguments below, as connect() is
-       a defined macro on some platforms and some compilers don't like to mix
-       #ifdefs with macro usage! (AmigaOS is one such platform) */
-#ifdef ENABLE_IPV6
-    rc = connect(sockfd, ai->ai_addr, ai->ai_addrlen);
-#else
-    rc = connect(sockfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr));
-#endif
-
-    if(-1 == rc) {
-      error = Curl_ourerrno();
-
-      switch (error) {
-      case EINPROGRESS:
-      case EWOULDBLOCK:
-#if defined(EAGAIN) && EAGAIN != EWOULDBLOCK
-        /* On some platforms EAGAIN and EWOULDBLOCK are the
-         * same value, and on others they are different, hence
-         * the odd #if
-         */
-      case EAGAIN:
-#endif
-        /* asynchronous connect, wait for connect or timeout */
-        if(data->state.used_interface == Curl_if_multi)
-          /* don't hang when doing multi */
-          timeout_ms = 0;
-        
-        rc = waitconnect(sockfd, timeout_ms);
-        break;
-      default:
-        /* unknown error, fallthrough and try another address! */
-        failf(data, "Failed to connect to %s IP number %d: %s",
-              hostname, aliasindex+1, Curl_strerror(conn,error));
-        break;
-      }
-    }
-
-    /* The '1 == rc' comes from the waitconnect(), and not from connect().
-       We can be sure of this since connect() cannot return 1. */
-    if((1 == rc) && (data->state.used_interface == Curl_if_multi)) {
-      /* Timeout when running the multi interface, we return here with a
-         CURLE_OK return code. */
-      rc = 0;
+    if(sockfd != CURL_SOCKET_BAD)
       break;
-    }
-      
-    if(0 == rc) {
-      if (verifyconnect(sockfd,NULL)) {
-        /* we are connected, awesome! */
-        *connected = TRUE; /* this is a true connect */
-        break;
-      }
-      /* nope, not connected for real */
-      rc = -1;
-    }
-    else
-      verifyconnect(sockfd,&error);    /* get non-blocking error */
-
-    /* connect failed or timed out */
-    sclose(sockfd);
-    sockfd = CURL_SOCKET_BAD;
 
     /* get a new timeout for next attempt */
     after = Curl_tvnow();
@@ -727,28 +819,26 @@ CURLcode Curl_connecthost(struct connectdata *conn,  /* context */
       return CURLE_OPERATION_TIMEOUTED;
     }
     before = after;
-  }
+  }  /* end of connect-to-each-address loop */
+
   if (sockfd == CURL_SOCKET_BAD) {
     /* no good connect was made */
-    *sockconn = -1;
-    failf(data, "Connect failed; %s", Curl_strerror(conn,error));
+    *sockconn = CURL_SOCKET_BAD;
+    failf(data, "couldn't connect to host");
     return CURLE_COULDNT_CONNECT;
   }
 
   /* leave the socket in non-blocking mode */
 
   /* store the address we use */
-  if(addr) {
-#ifdef ENABLE_IPV6
-    *addr = ai;
-#else
-    *addr = (struct in_addr *)remotehost->addr->h_addr_list[aliasindex];
-#endif
-  }
+  if(addr)
+    *addr = curr_addr;
 
   /* allow NULL-pointers to get passed in */
   if(sockconn)
     *sockconn = sockfd;    /* the socket descriptor we've connected */
 
+  data->info.numconnects++; /* to track the number of connections made */
+
   return CURLE_OK;
 }
