@@ -5,7 +5,7 @@
  *                            | (__| |_| |  _ <| |___
  *                             \___|\___/|_| \_\_____|
  *
- * Copyright (C) 1998 - 2004, Daniel Stenberg, <daniel@haxx.se>, et al.
+ * Copyright (C) 1998 - 2005, Daniel Stenberg, <daniel@haxx.se>, et al.
  *
  * This software is licensed as described in the file COPYING, which
  * you should have received as part of this distribution. The terms
@@ -18,7 +18,7 @@
  * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
  * KIND, either express or implied.
  *
- * $Id: transfer.c,v 1.250 2004/10/06 13:37:12 bagder Exp $
+ * $Id: transfer.c,v 1.290 2005/11/24 10:22:47 bagder Exp $
  ***************************************************************************/
 
 #include "setup.h"
@@ -32,7 +32,9 @@
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
+#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif
 
 #include <errno.h>
 
@@ -73,9 +75,6 @@
 #include <sys/select.h>
 #endif
 
-#ifndef HAVE_SELECT
-#error "We can't compile without select() support!"
-#endif
 #ifndef HAVE_SOCKET
 #error "We can't compile without socket() support!"
 #endif
@@ -95,12 +94,13 @@
 #include "http.h"
 #include "url.h"
 #include "getinfo.h"
-#include "ssluse.h"
+#include "sslgen.h"
 #include "http_digest.h"
 #include "http_ntlm.h"
 #include "http_negotiate.h"
 #include "share.h"
 #include "memory.h"
+#include "select.h"
 
 #define _MPRINTF_REPLACE /* use our functions only */
 #include <curl/mprintf.h>
@@ -116,10 +116,6 @@ enum {
   KEEP_WRITE
 };
 
-/* We keep this static and global since this is read-only and NEVER
-   changed. It should just remain a blanked-out timeout value. */
-static struct timeval notimeout={0,0};
-
 /*
  * This function will call the read callback to fill our buffer with data
  * to upload.
@@ -197,6 +193,69 @@ checkhttpprefix(struct SessionHandle *data,
   return FALSE;
 }
 
+/*
+ * Curl_readrewind() rewinds the read stream. This typically (so far) only
+ * used for HTTP POST/PUT with multi-pass authentication when a sending was
+ * denied and a resend is necessary.
+ */
+CURLcode Curl_readrewind(struct connectdata *conn)
+{
+  struct SessionHandle *data = conn->data;
+
+  conn->bits.rewindaftersend = FALSE; /* we rewind now */
+
+  /* We have sent away data. If not using CURLOPT_POSTFIELDS or
+     CURLOPT_HTTPPOST, call app to rewind
+  */
+  if(data->set.postfields ||
+     (data->set.httpreq == HTTPREQ_POST_FORM))
+    ; /* do nothing */
+  else {
+    if(data->set.ioctl) {
+      curlioerr err;
+
+      err = (data->set.ioctl) (data, CURLIOCMD_RESTARTREAD,
+                            data->set.ioctl_client);
+      infof(data, "the ioctl callback returned %d\n", (int)err);
+
+      if(err) {
+        /* FIXME: convert to a human readable error message */
+        failf(data, "ioctl callback returned error %d\n", (int)err);
+        return CURLE_SEND_FAIL_REWIND;
+      }
+    }
+    else {
+      /* If no CURLOPT_READFUNCTION is used, we know that we operate on a
+         given FILE * stream and we can actually attempt to rewind that
+         ourself with fseek() */
+      if(data->set.fread == (curl_read_callback)fread) {
+        if(-1 != fseek(data->set.in, 0, SEEK_SET))
+          /* successful rewind */
+          return CURLE_OK;
+      }
+
+      /* no callback set or failure aboe, makes us fail at once */
+      failf(data, "necessary data rewind wasn't possible\n");
+      return CURLE_SEND_FAIL_REWIND;
+    }
+  }
+  return CURLE_OK;
+}
+
+#ifdef USE_SSLEAY
+/* FIX: this is nasty OpenSSL-specific code that really shouldn't be here */
+static int data_pending(struct connectdata *conn)
+{
+  if(conn->ssl[FIRSTSOCKET].handle)
+    /* SSL is in use */
+    return SSL_pending(conn->ssl[FIRSTSOCKET].handle);
+
+  return 0; /* nothing */
+}
+#else
+/* non-SSL never have pending data */
+#define data_pending(x) 0
+#endif
 
 /*
  * Curl_readwrite() is the low-level function to be called when data is to
@@ -211,43 +270,32 @@ CURLcode Curl_readwrite(struct connectdata *conn,
   ssize_t nread; /* number of bytes read */
   int didwhat=0;
 
-  /* These two are used only if no other select() or _fdset() have been
-     invoked before this. This typicly happens if you use the multi interface
-     and call curl_multi_perform() without calling curl_multi_fdset()
-     first. */
-  fd_set extrareadfd;
-  fd_set extrawritefd;
+  curl_socket_t fd_read;
+  curl_socket_t fd_write;
+  curl_socket_t select_res;
 
-  fd_set *readfdp = k->readfdp;
-  fd_set *writefdp = k->writefdp;
   curl_off_t contentlength;
 
-  if((k->keepon & KEEP_READ) && !readfdp) {
-    /* reading is requested, but no socket descriptor pointer was set */
-    FD_ZERO(&extrareadfd);
-    FD_SET(conn->sockfd, &extrareadfd);
-    readfdp = &extrareadfd;
+  if(k->keepon & KEEP_READ)
+    fd_read = conn->sockfd;
+  else
+    fd_read = CURL_SOCKET_BAD;
 
-    /* no write, no exceptions, no timeout */
-    select(conn->sockfd+1, readfdp, NULL, NULL, &notimeout);
-  }
-  if((k->keepon & KEEP_WRITE) && !writefdp) {
-    /* writing is requested, but no socket descriptor pointer was set */
-    FD_ZERO(&extrawritefd);
-    FD_SET(conn->writesockfd, &extrawritefd);
-    writefdp = &extrawritefd;
+  if(k->keepon & KEEP_WRITE)
+    fd_write = conn->writesockfd;
+  else
+    fd_write = CURL_SOCKET_BAD;
 
-    /* no read, no exceptions, no timeout */
-    select(conn->writesockfd+1, NULL, writefdp, NULL, &notimeout);
+  select_res = Curl_select(fd_read, fd_write, 0);
+  if(select_res == CSELECT_ERR) {
+    failf(data, "select/poll returned error");
+    return CURLE_SEND_ERROR;
   }
 
   do {
     /* If we still have reading to do, we check if we have a readable
-       socket. Sometimes the reafdp is NULL, if no fd_set was done using
-       the multi interface and then we can do nothing but to attempt a
-       read to be sure. */
-    if((k->keepon & KEEP_READ) &&
-       (!readfdp || FD_ISSET(conn->sockfd, readfdp))) {
+       socket. */
+    if((k->keepon & KEEP_READ) && (select_res & CSELECT_IN)) {
 
       bool is_empty_data = FALSE;
 
@@ -255,7 +303,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
          read or we get a EWOULDBLOCK */
       do {
         size_t buffersize = data->set.buffer_size?
-          data->set.buffer_size:BUFSIZE -1;
+          data->set.buffer_size:BUFSIZE;
 
         /* receive data from the network! */
         int readrc = Curl_read(conn, conn->sockfd, k->buf, buffersize, &nread);
@@ -289,7 +337,6 @@ CURLcode Curl_readwrite(struct connectdata *conn,
            we bail out from this! */
         else if (0 >= nread) {
           k->keepon &= ~KEEP_READ;
-          FD_ZERO(&k->rkeepfd);
           break;
         }
 
@@ -313,7 +360,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
             /* str_start is start of line within buf */
             k->str_start = k->str;
 
-            k->end_ptr = strchr (k->str_start, '\n');
+            k->end_ptr = memchr(k->str_start, '\n', nread);
 
             if (!k->end_ptr) {
               /* Not a complete header line within buffer, append the data to
@@ -352,7 +399,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
 
             /* decrease the size of the remaining (supposed) header line */
             rest_length = (k->end_ptr - k->str)+1;
-            nread -= rest_length;
+            nread -= (ssize_t)rest_length;
 
             k->str = k->end_ptr + 1; /* move past new line */
 
@@ -381,7 +428,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
             }
 
             /* copy to end of line */
-            strncpy (k->hbufp, k->str_start, full_length);
+            memcpy(k->hbufp, k->str_start, full_length);
             k->hbufp += full_length;
             k->hbuflen += full_length;
             *k->hbufp = 0;
@@ -434,9 +481,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                 if (k->write_after_100_header) {
 
                   k->write_after_100_header = FALSE;
-                  FD_SET (conn->writesockfd, &k->writefd); /* write */
                   k->keepon |= KEEP_WRITE;
-                  k->wkeepfd = k->writefd;
                 }
               }
               else
@@ -451,7 +496,6 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                  */
                 k->write_after_100_header = FALSE;
                 k->keepon &= ~KEEP_WRITE;
-                FD_ZERO(&k->wkeepfd);
               }
 
 #ifndef CURL_DISABLE_HTTP
@@ -480,14 +524,13 @@ CURLcode Curl_readwrite(struct connectdata *conn,
               if(result)
                 return result;
 
-              data->info.header_size += headerlen;
-              conn->headerbytecount += headerlen;
+              data->info.header_size += (long)headerlen;
+              conn->headerbytecount += (long)headerlen;
 
               conn->deductheadercount =
                 (100 == k->httpcode)?conn->headerbytecount:0;
 
               if (conn->resume_from &&
-                  !k->content_range &&
                   (data->set.httpreq==HTTPREQ_GET) &&
                   (k->httpcode == 416)) {
                 /* "Requested Range Not Satisfiable" */
@@ -503,6 +546,13 @@ CURLcode Curl_readwrite(struct connectdata *conn,
 
                 if(result)
                   return result;
+
+                if(conn->bits.rewindaftersend) {
+                  /* We rewind after a complete send, so thus we continue
+                     sending now */
+                  infof(data, "Keep sending data to get tossed away!\n");
+                  k->keepon |= KEEP_WRITE;
+                }
               }
 #endif   /* CURL_DISABLE_HTTP */
 
@@ -549,7 +599,6 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                 if(stop_reading) {
                   /* we make sure that this socket isn't read more now */
                   k->keepon &= ~KEEP_READ;
-                  FD_ZERO(&k->rkeepfd);
                 }
 
                 break;          /* exit header line loop */
@@ -613,10 +662,19 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                     (k->httpcode >= 400) &&
                     (k->httpcode != 401) &&
                     (k->httpcode != 407)) {
-                  /* serious error, go home! */
-                  failf (data, "The requested URL returned error: %d",
-                         k->httpcode);
-                  return CURLE_HTTP_RETURNED_ERROR;
+
+                  if (conn->resume_from &&
+                      (data->set.httpreq==HTTPREQ_GET) &&
+                      (k->httpcode == 416)) {
+                    /* "Requested Range Not Satisfiable", just proceed and
+                       pretend this is no error */
+                  }
+                  else {
+                    /* serious error, go home! */
+                    failf (data, "The requested URL returned error: %d",
+                           k->httpcode);
+                    return CURLE_HTTP_RETURNED_ERROR;
+                  }
                 }
 
                 if(k->httpversion == 10)
@@ -643,6 +701,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                    * fields.  */
                   conn->size=0;
                   conn->maxdownload=0;
+                  k->ignorecl = TRUE; /* ignore Content-Length headers */
                   break;
                 default:
                   /* nothing */
@@ -659,7 +718,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                the header completely if we get a 416 response as then we're
                resuming a document that we don't get, and this header contains
                info about the true size of the document we didn't get now. */
-            if ((k->httpcode != 416) &&
+            if (!k->ignorecl && !data->set.ignorecl &&
                 checkprefix("Content-Length:", k->p)) {
               contentlength = curlx_strtoofft(k->p+15, NULL, 10);
               if (data->set.max_filesize &&
@@ -687,7 +746,8 @@ CURLcode Curl_readwrite(struct connectdata *conn,
               /* Find the first non-space letter */
               for(start=k->p+13;
                   *start && isspace((int)*start);
-                  start++);
+                  start++)
+                ;  /* empty loop */
 
               end = strchr(start, '\r');
               if(!end)
@@ -695,7 +755,8 @@ CURLcode Curl_readwrite(struct connectdata *conn,
 
               if(end) {
                 /* skip all trailing space letters */
-                for(; isspace((int)*end) && (end > start); end--);
+                for(; isspace((int)*end) && (end > start); end--)
+                  ;  /* empty loop */
 
                 /* get length of the type */
                 len = end-start+1;
@@ -772,6 +833,20 @@ CURLcode Curl_readwrite(struct connectdata *conn,
               /* init our chunky engine */
               Curl_httpchunk_init(conn);
             }
+
+            else if (checkprefix("Trailer:", k->p) ||
+                     checkprefix("Trailers:", k->p)) {
+              /*
+               * This test helps Curl_httpchunk_read() to determine to look
+               * for well formed trailers after the zero chunksize record. In
+               * this case a CRLF is required after the zero chunksize record
+               * when no trailers are sent, or after the last trailer record.
+               *
+               * It seems both Trailer: and Trailers: occur in the wild.
+               */
+              conn->bits.trailerHdrPresent = TRUE;
+            }
+
             else if (checkprefix("Content-Encoding:", k->p) &&
                      data->set.encoding) {
               /*
@@ -786,7 +861,8 @@ CURLcode Curl_readwrite(struct connectdata *conn,
               /* Find the first non-space letter */
               for(start=k->p+17;
                   *start && isspace((int)*start);
-                  start++);
+                  start++)
+                ;  /* empty loop */
 
               /* Record the content-encoding for later use */
               if (checkprefix("identity", start))
@@ -804,11 +880,11 @@ CURLcode Curl_readwrite(struct connectdata *conn,
               /* Content-Range: bytes [num]-
                  Content-Range: bytes: [num]-
 
-                 The second format was added August 1st 2000 by Igor
-                 Khristophorov since Sun's webserver JavaWebServer/1.1.1
-                 obviously sends the header this way! :-( */
+                 The second format was added since Sun's webserver
+                 JavaWebServer/1.1.1 obviously sends the header this way!
+              */
 
-              char *ptr = strstr(k->p, "bytes");
+              char *ptr = Curl_strcasestr(k->p, "bytes");
               ptr+=5;
 
               if(*ptr == ':')
@@ -821,6 +897,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                 /* we asked for a resume and we got it */
                 k->content_range = TRUE;
             }
+#if !defined(CURL_DISABLE_COOKIES)
             else if(data->cookies &&
                     checkprefix("Set-Cookie:", k->p)) {
               Curl_share_lock(data, CURL_LOCK_DATA_COOKIE,
@@ -834,13 +911,14 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                               conn->path);
               Curl_share_unlock(data, CURL_LOCK_DATA_COOKIE);
             }
+#endif
             else if(checkprefix("Last-Modified:", k->p) &&
                     (data->set.timecondition || data->set.get_filetime) ) {
               time_t secs=time(NULL);
               k->timeofdoc = curl_getdate(k->p+strlen("Last-Modified:"),
                                           &secs);
               if(data->set.get_filetime)
-                data->info.filetime = k->timeofdoc;
+                data->info.filetime = (long)k->timeofdoc;
             }
             else if((checkprefix("WWW-Authenticate:", k->p) &&
                      (401 == k->httpcode)) ||
@@ -897,14 +975,14 @@ CURLcode Curl_readwrite(struct connectdata *conn,
 
             if(data->set.verbose)
               Curl_debug(data, CURLINFO_HEADER_IN,
-                         k->p, k->hbuflen, conn->host.dispname);
+                         k->p, k->hbuflen, conn);
 
             result = Curl_client_write(data, writetype, k->p, k->hbuflen);
             if(result)
               return result;
 
-            data->info.header_size += k->hbuflen;
-            conn->headerbytecount += k->hbuflen;
+            data->info.header_size += (long)k->hbuflen;
+            conn->headerbytecount += (long)k->hbuflen;
 
             /* reset hbufp pointer && hbuflen */
             k->hbufp = data->state.headerbuff;
@@ -938,7 +1016,6 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                   /* Abort after the headers if "follow Location" is set
                      and we're set to close anyway. */
                   k->keepon &= ~KEEP_READ;
-                  FD_ZERO(&k->rkeepfd);
                   *done = TRUE;
                   return CURLE_OK;
                 }
@@ -954,8 +1031,8 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                 /* we wanted to resume a download, although the server doesn't
                  * seem to support this and we did this with a GET (if it
                  * wasn't a GET we did a POST or PUT resume) */
-                failf (data, "HTTP server doesn't seem to support "
-                       "byte ranges. Cannot resume.");
+                failf(data, "HTTP server doesn't seem to support "
+                      "byte ranges. Cannot resume.");
                 return CURLE_HTTP_RANGE_ERROR;
               }
 
@@ -995,14 +1072,12 @@ CURLcode Curl_readwrite(struct connectdata *conn,
           if(data->set.verbose) {
             if(k->badheader) {
               Curl_debug(data, CURLINFO_DATA_IN, data->state.headerbuff,
-                         k->hbuflen, conn->host.dispname);
+                         k->hbuflen, conn);
               if(k->badheader == HEADER_PARTHEADER)
-                Curl_debug(data, CURLINFO_DATA_IN, k->str, nread,
-                           conn->host.dispname);
+                Curl_debug(data, CURLINFO_DATA_IN, k->str, nread, conn);
             }
             else
-              Curl_debug(data, CURLINFO_DATA_IN, k->str, nread,
-                         conn->host.dispname);
+              Curl_debug(data, CURLINFO_DATA_IN, k->str, nread, conn);
           }
 
 #ifndef CURL_DISABLE_HTTP
@@ -1013,6 +1088,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
              * the name says read, this function both reads and writes away
              * the data. The returned 'nread' holds the number of actual
              * data it wrote to the client.  */
+
             CHUNKcode res =
               Curl_httpchunk_read(conn, k->str, nread, &nread);
 
@@ -1027,7 +1103,6 @@ CURLcode Curl_readwrite(struct connectdata *conn,
             else if(CHUNKE_STOP == res) {
               /* we're done reading chunks! */
               k->keepon &= ~KEEP_READ; /* read no more */
-              FD_ZERO(&k->rkeepfd);
 
               /* There are now possibly N number of bytes at the end of the
                  str buffer that weren't written to the client, but we don't
@@ -1044,7 +1119,6 @@ CURLcode Curl_readwrite(struct connectdata *conn,
               nread = 0;
 
             k->keepon &= ~KEEP_READ; /* we're done reading */
-            FD_ZERO(&k->rkeepfd);
           }
 
           k->bytecount += nread;
@@ -1060,6 +1134,8 @@ CURLcode Curl_readwrite(struct connectdata *conn,
               result = Curl_client_write(data, CLIENTWRITE_BODY,
                                          data->state.headerbuff,
                                          k->hbuflen);
+              if(result)
+                return result;
             }
             if(k->badheader < HEADER_ALLBAD) {
               /* This switch handles various content encodings. If there's an
@@ -1112,19 +1188,15 @@ CURLcode Curl_readwrite(struct connectdata *conn,
           /* if we received nothing, the server closed the connection and we
              are done */
           k->keepon &= ~KEEP_READ;
-          FD_ZERO(&k->rkeepfd);
         }
 
-      } while(0);
+      } while(data_pending(conn));
 
     } /* if( read from socket ) */
 
     /* If we still have writing to do, we check if we have a writable
-       socket. Sometimes the writefdp is NULL, if no fd_set was done using
-       the multi interface and then we can do nothing but to attempt a
-       write to be sure. */
-    if((k->keepon & KEEP_WRITE) &&
-       (!writefdp || FD_ISSET(conn->writesockfd, writefdp)) ) {
+       socket. */
+    if((k->keepon & KEEP_WRITE) && (select_res & CSELECT_OUT)) {
       /* write */
 
       int i, si;
@@ -1160,8 +1232,6 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                  go into the Expect: 100 state and await such a header */
               k->wait100_after_headers = FALSE; /* headers sent */
               k->write_after_100_header = TRUE; /* wait for the header */
-              FD_ZERO (&k->writefd);            /* clear it */
-              k->wkeepfd = k->writefd;          /* set the keeper variable */
               k->keepon &= ~KEEP_WRITE;         /* disable writing */
               k->start100 = Curl_tvnow();       /* timeout count starts now */
               didwhat &= ~KEEP_WRITE;  /* we didn't write anything actually */
@@ -1182,8 +1252,13 @@ CURLcode Curl_readwrite(struct connectdata *conn,
           if (nread<=0) {
             /* done */
             k->keepon &= ~KEEP_WRITE; /* we're done writing */
-            FD_ZERO(&k->wkeepfd);
             writedone = TRUE;
+
+            if(conn->bits.rewindaftersend) {
+              result = Curl_readrewind(conn);
+              if(result)
+                return result;
+            }
             break;
           }
 
@@ -1236,7 +1311,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
         if(data->set.verbose)
           /* show the data before we change the pointer upload_fromhere */
           Curl_debug(data, CURLINFO_DATA_OUT, conn->upload_fromhere,
-                     bytes_written, conn->host.dispname);
+                     bytes_written, conn);
 
         if(conn->upload_present != bytes_written) {
           /* we only wrote a part of the buffer (if anything), deal with it! */
@@ -1258,7 +1333,6 @@ CURLcode Curl_readwrite(struct connectdata *conn,
           if(k->upload_done) {
             /* switch off writing, we're done! */
             k->keepon &= ~KEEP_WRITE; /* we're done writing */
-            FD_ZERO(&k->wkeepfd);
             writedone = TRUE;
           }
         }
@@ -1300,9 +1374,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
       if(ms > CURL_TIMEOUT_EXPECT_100) {
         /* we've waited long enough, continue anyway */
         k->write_after_100_header = FALSE;
-        FD_SET (conn->writesockfd, &k->writefd); /* write socket */
         k->keepon |= KEEP_WRITE;
-        k->wkeepfd = k->writefd;
       }
     }
   }
@@ -1336,9 +1408,19 @@ CURLcode Curl_readwrite(struct connectdata *conn,
             conn->size - k->bytecount);
       return CURLE_PARTIAL_FILE;
     }
-    else if(conn->bits.chunk && conn->proto.http->chunk.datasize) {
-      failf(data, "transfer closed with at least %d bytes remaining",
-            conn->proto.http->chunk.datasize);
+    else if(!(conn->bits.no_body) &&
+            conn->bits.chunk &&
+            (conn->proto.http->chunk.state != CHUNK_STOP)) {
+      /*
+       * In chunked mode, return an error if the connection is closed prior to
+       * the empty (terminiating) chunk is read.
+       *
+       * The condition above used to check for
+       * conn->proto.http->chunk.datasize != 0 which is true after reading
+       * *any* chunk, not just the empty chunk.
+       *
+       */
+      failf(data, "transfer closed with outstanding read data remaining");
       return CURLE_PARTIAL_FILE;
     }
     if(Curl_pgrsUpdate(conn))
@@ -1392,13 +1474,10 @@ CURLcode Curl_readwrite_init(struct connectdata *conn)
   /* we want header and/or body, if neither then don't do this! */
   if(conn->bits.getheader || !conn->bits.no_body) {
 
-    FD_ZERO (&k->readfd);               /* clear it */
     if(conn->sockfd != CURL_SOCKET_BAD) {
-      FD_SET (conn->sockfd, &k->readfd); /* read socket */
       k->keepon |= KEEP_READ;
     }
 
-    FD_ZERO (&k->writefd);              /* clear it */
     if(conn->writesockfd != CURL_SOCKET_BAD) {
       /* HTTP 1.1 magic:
 
@@ -1420,16 +1499,9 @@ CURLcode Curl_readwrite_init(struct connectdata *conn)
           /* when we've sent off the rest of the headers, we must await a
              100-continue */
           k->wait100_after_headers = TRUE;
-        FD_SET (conn->writesockfd, &k->writefd); /* write socket */
         k->keepon |= KEEP_WRITE;
       }
     }
-
-    /* get these in backup variables to be able to restore them on each lap in
-       the select() loop */
-    k->rkeepfd = k->readfd;
-    k->wkeepfd = k->writefd;
-
   }
 
   return CURLE_OK;
@@ -1451,8 +1523,7 @@ void Curl_single_fdset(struct connectdata *conn,
   *max_fd = -1; /* init */
   if(conn->keep.keepon & KEEP_READ) {
     FD_SET(conn->sockfd, read_fd_set);
-    *max_fd = conn->sockfd;
-    conn->keep.readfdp = read_fd_set; /* store the address of the set */
+    *max_fd = (int)conn->sockfd;
   }
   if(conn->keep.keepon & KEEP_WRITE) {
     FD_SET(conn->writesockfd, write_fd_set);
@@ -1460,8 +1531,7 @@ void Curl_single_fdset(struct connectdata *conn,
     /* since sockets are curl_socket_t nowadays, we typecast it to int here
        to compare it nicely */
     if((int)conn->writesockfd > *max_fd)
-      *max_fd = conn->writesockfd;
-    conn->keep.writefdp = write_fd_set; /* store the address of the set */
+      *max_fd = (int)conn->writesockfd;
   }
   /* we don't use exceptions, only touch that one to prevent compiler
      warnings! */
@@ -1504,17 +1574,24 @@ Transfer(struct connectdata *conn)
   if(!conn->bits.getheader && conn->bits.no_body)
     return CURLE_OK;
 
-  k->writefdp = &k->writefd; /* store the address of the set */
-  k->readfdp = &k->readfd;   /* store the address of the set */
-
   while (!done) {
-    struct timeval interval;
-    k->readfd = k->rkeepfd;  /* set these every lap in the loop */
-    k->writefd = k->wkeepfd;
-    interval.tv_sec = 1;
-    interval.tv_usec = 0;
+    curl_socket_t fd_read;
+    curl_socket_t fd_write;
+    int interval_ms;
 
-    switch (select (k->maxfd, k->readfdp, k->writefdp, NULL, &interval)) {
+    interval_ms = 1 * 1000;
+
+    if(k->keepon & KEEP_READ)
+      fd_read = conn->sockfd;
+    else
+      fd_read = CURL_SOCKET_BAD;
+
+    if(k->keepon & KEEP_WRITE)
+      fd_write = conn->writesockfd;
+    else
+      fd_write = CURL_SOCKET_BAD;
+
+    switch (Curl_select(fd_read, fd_write, interval_ms)) {
     case -1: /* select() error, stop reading */
 #ifdef EINTR
       /* The EINTR is not serious, and it seems you might get this more
@@ -1544,20 +1621,19 @@ Transfer(struct connectdata *conn)
  */
 CURLcode Curl_pretransfer(struct SessionHandle *data)
 {
-  if(!data->change.url)
+  CURLcode res;
+  if(!data->change.url) {
     /* we can't do anything wihout URL */
+    failf(data, "No URL set!\n");
     return CURLE_URL_MALFORMAT;
-
-#ifdef USE_SSLEAY
-  {
-    /* Init the SSL session ID cache here. We do it here since we want to do
-       it after the *_setopt() calls (that could change the size of the cache)
-       but before any transfer takes place. */
-    CURLcode res = Curl_SSL_InitSessions(data, data->set.ssl.numsessions);
-    if(res)
-      return res;
   }
-#endif
+
+  /* Init the SSL session ID cache here. We do it here since we want to do it
+     after the *_setopt() calls (that could change the size of the cache) but
+     before any transfer takes place. */
+  res = Curl_ssl_initsessions(data, data->set.ssl.numsessions);
+  if(res)
+    return res;
 
   data->set.followlocation=0; /* reset the location-follow counter */
   data->state.this_is_a_follow = FALSE; /* reset this */
@@ -1567,25 +1643,9 @@ CURLcode Curl_pretransfer(struct SessionHandle *data)
   data->state.authhost.want = data->set.httpauth;
   data->state.authproxy.want = data->set.proxyauth;
 
-#ifndef CURL_DISABLE_HTTP
-  /* If there was a list of cookie files to read and we haven't done it before,
-     do it now! */
-  if(data->change.cookielist) {
-    struct curl_slist *list = data->change.cookielist;
-    Curl_share_lock(data, CURL_LOCK_DATA_COOKIE, CURL_LOCK_ACCESS_SINGLE);
-    while(list) {
-      data->cookies = Curl_cookie_init(data,
-                                       list->data,
-                                       data->cookies,
-                                       data->set.cookiesession);
-      list = list->next;
-    }
-    Curl_share_unlock(data, CURL_LOCK_DATA_COOKIE);
-    curl_slist_free_all(data->change.cookielist); /* clean up list */
-    data->change.cookielist = NULL; /* don't do this again! */
-  }
-#endif   /* CURL_DISABLE_HTTP */
-
+  /* If there is a list of cookie files to read, do it now! */
+  if(data->change.cookielist)
+    Curl_cookie_loadfiles(data);
 
  /* Allow data->set.use_port to set which port to use. This needs to be
   * disabled for example when we follow Location: headers to URLs using
@@ -1619,6 +1679,11 @@ CURLcode Curl_posttransfer(struct SessionHandle *data)
   (void)data; /* unused parameter */
 #endif
 
+  if(!(data->progress.flags & PGRS_HIDE) &&
+     !data->progress.callback)
+    /* only output if we don't use a progress callback and we're not hidden */
+    fprintf(data->set.err, "\n");
+
   return CURLE_OK;
 }
 
@@ -1688,9 +1753,11 @@ static void strcpy_url(char *output, char *url)
  * as given by the remote server and set up the new URL to request.
  */
 CURLcode Curl_follow(struct SessionHandle *data,
-                     char *newurl) /* this 'newurl' is the Location: string,
+                     char *newurl, /* this 'newurl' is the Location: string,
                                       and it must be malloc()ed before passed
                                       here */
+                     bool retry) /* set TRUE if this is a request retry as
+                                    opposed to a real redirect following */
 {
   /* Location: redirect */
   char prot[16]; /* URL protocol string storage */
@@ -1698,14 +1765,15 @@ CURLcode Curl_follow(struct SessionHandle *data,
   size_t newlen;
   char *newest;
 
-  if (data->set.maxredirs &&
+  if ((data->set.maxredirs != -1) &&
       (data->set.followlocation >= data->set.maxredirs)) {
     failf(data,"Maximum (%d) redirects followed", data->set.maxredirs);
     return CURLE_TOO_MANY_REDIRECTS;
   }
 
-  /* mark the next request as a followed location: */
-  data->state.this_is_a_follow = TRUE;
+  if(!retry)
+    /* mark the next request as a followed location: */
+    data->state.this_is_a_follow = TRUE;
 
   data->set.followlocation++; /* count location-followers */
 
@@ -1986,9 +2054,11 @@ Curl_connect_host(struct SessionHandle *data,
 
   do {
     bool async;
+    bool protocol_done=TRUE; /* will be TRUE always since this is only used
+                                within the easy interface */
     Curl_pgrsTime(data, TIMER_STARTSINGLE);
     data->change.url_changed = FALSE;
-    res = Curl_connect(data, conn, &async);
+    res = Curl_connect(data, conn, &async, &protocol_done);
 
     if((CURLE_OK == res) && async) {
       /* Now, if async is TRUE here, we need to wait for the name
@@ -1996,7 +2066,10 @@ Curl_connect_host(struct SessionHandle *data,
       res = Curl_wait_for_resolv(*conn, NULL);
       if(CURLE_OK == res)
         /* Resolved, continue with the connection */
-        res = Curl_async_resolved(*conn);
+        res = Curl_async_resolved(*conn, &protocol_done);
+      else
+        /* if we can't resolve, we kill this "connection" now */
+        (void)Curl_disconnect(*conn);
     }
     if(res)
       break;
@@ -2009,7 +2082,7 @@ Curl_connect_host(struct SessionHandle *data,
       res = Curl_done(conn, res);
       if(CURLE_OK == res) {
         char *gotourl = strdup(data->change.url);
-        res = Curl_follow(data, gotourl);
+        res = Curl_follow(data, gotourl, FALSE);
         if(res)
           free(gotourl);
       }
@@ -2019,7 +2092,32 @@ Curl_connect_host(struct SessionHandle *data,
   return res;
 }
 
+/* Returns TRUE and sets '*url' if a request retry is wanted */
+bool Curl_retry_request(struct connectdata *conn,
+                        char **url)
+{
+  bool retry = FALSE;
 
+  if((conn->keep.bytecount+conn->headerbytecount == 0) &&
+     conn->bits.reuse) {
+    /* We got no data and we attempted to re-use a connection. This might
+       happen if the connection was left alive when we were done using it
+       before, but that was closed when we wanted to read from it again. Bad
+       luck. Retry the same request on a fresh connect! */
+    infof(conn->data, "Connection died, retrying a fresh connect\n");
+    *url = strdup(conn->data->change.url);
+
+    conn->bits.close = TRUE; /* close this connection */
+    conn->bits.retry = TRUE; /* mark this as a connection we're about
+                                to retry. Marking it this way should
+                                prevent i.e HTTP transfers to return
+                                error just because nothing has been
+                                transfered! */
+    retry = TRUE;
+  }
+
+  return retry;
+}
 
 /*
  * Curl_perform() is the internal high-level function that gets called by the
@@ -2032,6 +2130,7 @@ CURLcode Curl_perform(struct SessionHandle *data)
   CURLcode res2;
   struct connectdata *conn=NULL;
   char *newurl = NULL; /* possibly a new URL to follow to! */
+  bool retry = FALSE;
 
   data->state.used_interface = Curl_if_easy;
 
@@ -2050,42 +2149,26 @@ CURLcode Curl_perform(struct SessionHandle *data)
     res = Curl_connect_host(data, &conn);   /* primary connection */
 
     if(res == CURLE_OK) {
-      if (data->set.source_host) /* 3rd party transfer */
-        res = Curl_pretransfersec(conn);
+      if (data->set.source_url) /* 3rd party transfer */
+        res = Curl_second_connect(conn);
       else
         conn->sec_conn = NULL;
     }
 
     if(res == CURLE_OK) {
-
-      res = Curl_do(&conn);
+      bool do_done;
+      res = Curl_do(&conn, &do_done);
 
       /* for non 3rd party transfer only */
-      if(res == CURLE_OK && !data->set.source_host) {
+      if(res == CURLE_OK && !data->set.source_url) {
         res = Transfer(conn); /* now fetch that URL please */
         if(res == CURLE_OK) {
+          retry = Curl_retry_request(conn, &newurl);
 
-          if((conn->keep.bytecount+conn->headerbytecount == 0) &&
-             conn->bits.reuse) {
-            /* We got no data and we attempted to re-use a connection. This
-               might happen if the connection was left alive when we were done
-               using it before, but that was closed when we wanted to read
-               from it again. Bad luck. Retry the same request on a fresh
-               connect! */
-            infof(data, "Connection died, retrying a fresh connect\n");
-            newurl = strdup(conn->data->change.url);
-
-            conn->bits.close = TRUE; /* close this connection */
-            conn->bits.retry = TRUE; /* mark this as a connection we're about
-                                        to retry. Marking it this way should
-                                        prevent i.e HTTP transfers to return
-                                        error just because nothing has been
-                                        transfered! */
-          }
-          else
+          if(!retry)
             /*
-             * We must duplicate the new URL here as the connection data
-             * may be free()ed in the Curl_done() function.
+             * We must duplicate the new URL here as the connection data may
+             * be free()ed in the Curl_done() function.
              */
             newurl = conn->newurl?strdup(conn->newurl):NULL;
         }
@@ -2120,7 +2203,7 @@ CURLcode Curl_perform(struct SessionHandle *data)
        */
 
       if((res == CURLE_OK) && newurl) {
-        res = Curl_follow(data, newurl);
+        res = Curl_follow(data, newurl, retry);
         if(CURLE_OK == res) {
           newurl = NULL;
           continue;
@@ -2134,6 +2217,19 @@ CURLcode Curl_perform(struct SessionHandle *data)
   if(newurl)
     free(newurl);
 
+  if(res && !data->state.errorbuf) {
+    /*
+     * As an extra precaution: if no error string has been set and there was
+     * an error, use the strerror() string or if things are so bad that not
+     * even that is good, set a bad string that mentions the error code.
+     */
+    const char *str = curl_easy_strerror(res);
+    if(!str)
+      failf(data, "unspecified error %d", (int)res);
+    else
+      failf(data, "%s", str);
+  }
+
   /* run post-transfer uncondionally, but don't clobber the return code if
      we already have an error code recorder */
   res2 = Curl_posttransfer(data);
@@ -2181,35 +2277,45 @@ Curl_Transfer(struct connectdata *c_conn, /* connection data */
 }
 
 /*
- * Curl_pretransfersec() prepares the secondary connection (used for 3rd party
+ * Curl_second_connect() makes the secondary connection (used for 3rd party
  * FTP transfers).
  */
-CURLcode Curl_pretransfersec(struct connectdata *conn)
+CURLcode Curl_second_connect(struct connectdata *conn)
 {
   CURLcode status = CURLE_OK;
   struct SessionHandle *data = conn->data;
   struct connectdata *sec_conn = NULL;   /* secondary connection */
-
-  /* update data with source host options */
-  char *url = aprintf( "%s://%s/", conn->protostr, data->set.source_host);
-
-  if(!url)
-    return CURLE_OUT_OF_MEMORY;
+  bool backup_reuse_fresh = data->set.reuse_fresh;
+  char *backup_userpwd = data->set.userpwd;
 
   if(data->change.url_alloc)
     free(data->change.url);
 
-  data->change.url_alloc = TRUE;
-  data->change.url = url;
-  data->set.ftpport = data->set.source_port;
+  data->change.url_alloc = FALSE;
+  data->change.url = data->set.source_url;
+
+  /* We must never actually alter 'data->set' properties, so we restore the
+     backed up values afterwards! */
+
+#if 0
+  /* if both remote hosts are the same host - create new connection */
+  if (strequal(conn->host.dispname, data->set.source_host))
+#endif
+    data->set.reuse_fresh = TRUE;
+
   data->set.userpwd = data->set.source_userpwd;
 
   /* secondary connection */
   status = Curl_connect_host(data, &sec_conn);
   if(CURLE_OK == status) {
+    sec_conn->sec_conn = NULL;  /* important if re-using existing connection
+                                   to prevent loop */
     sec_conn->data = data;
     conn->sec_conn = sec_conn;
   }
 
+  data->set.reuse_fresh = backup_reuse_fresh;
+  data->set.userpwd = backup_userpwd;
+
   return status;
 }
