@@ -1,16 +1,16 @@
 /***************************************************************************
- *                                  _   _ ____  _     
- *  Project                     ___| | | |  _ \| |    
- *                             / __| | | | |_) | |    
- *                            | (__| |_| |  _ <| |___ 
+ *                                  _   _ ____  _
+ *  Project                     ___| | | |  _ \| |
+ *                             / __| | | | |_) | |
+ *                            | (__| |_| |  _ <| |___
  *                             \___|\___/|_| \_\_____|
  *
- * Copyright (C) 1998 - 2004, Daniel Stenberg, <daniel@haxx.se>, et al.
+ * Copyright (C) 1998 - 2005, Daniel Stenberg, <daniel@haxx.se>, et al.
  *
  * This software is licensed as described in the file COPYING, which
  * you should have received as part of this distribution. The terms
  * are also available at http://curl.haxx.se/docs/copyright.html.
- * 
+ *
  * You may opt to use, copy, modify, merge, publish, distribute and/or sell
  * copies of the Software, and permit persons to whom the Software is
  * furnished to do so, under the terms of the COPYING file.
@@ -18,7 +18,7 @@
  * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
  * KIND, either express or implied.
  *
- * $Id: urlglob.c,v 1.34 2004/05/12 08:22:04 bagder Exp $
+ * $Id: urlglob.c,v 1.41 2005/11/10 22:11:02 bagder Exp $
  ***************************************************************************/
 
 /* client-local setup.h */
@@ -73,20 +73,22 @@ static GlobCode glob_set(URLGlob *glob, char *pattern,
   ++glob->size;
 
   while (1) {
+    bool skip;
+
     switch (*pattern) {
-    case '\0':			/* URL ended while set was still open */
+    case '\0':                  /* URL ended while set was still open */
       snprintf(glob->errormsg, sizeof(glob->errormsg),
                "unmatched brace at pos %d\n", (int)pos);
       return GLOB_ERROR;
 
     case '{':
-    case '[':			/* no nested expressions at this time */
+    case '[':                   /* no nested expressions at this time */
       snprintf(glob->errormsg, sizeof(glob->errormsg),
                "nested braces not supported at pos %d\n", (int)pos);
       return GLOB_ERROR;
 
     case ',':
-    case '}':				/* set element completed */
+    case '}':                           /* set element completed */
       *buf = '\0';
       pat->content.Set.elements =
         realloc(pat->content.Set.elements,
@@ -103,10 +105,10 @@ static GlobCode glob_set(URLGlob *glob, char *pattern,
         /* entire set pattern completed */
         int wordamount;
 
-	/* always check for a literal (may be "") between patterns */
+        /* always check for a literal (may be "") between patterns */
         if(GLOB_ERROR == glob_word(glob, ++pattern, ++pos, &wordamount))
           wordamount=1;
-	*amount = pat->content.Set.size * wordamount;
+        *amount = pat->content.Set.size * wordamount;
 
         return GLOB_OK;
       }
@@ -116,29 +118,44 @@ static GlobCode glob_set(URLGlob *glob, char *pattern,
       ++pos;
       break;
 
-    case ']':				/* illegal closing bracket */
-      snprintf(glob->errormsg, sizeof(glob->errormsg), 
+    case ']':                           /* illegal closing bracket */
+      snprintf(glob->errormsg, sizeof(glob->errormsg),
                "illegal pattern at pos %d\n", (int)pos);
       return GLOB_ERROR;
 
-    case '\\':				/* escaped character, skip '\' */
-      if (*(buf+1) == '\0') {		/* but no escaping of '\0'! */
-        snprintf(glob->errormsg, sizeof(glob->errormsg), 
-                 "illegal pattern at pos %d\n", (int)pos);
-	return GLOB_ERROR;
+    case '\\':                          /* escaped character, skip '\' */
+      switch(pattern[1]) {
+      case '[':
+      case ']':
+      case '{':
+      case '}':
+      case ',':
+        skip = TRUE;
+        break;
+      default:
+        skip = FALSE;
+        break;
       }
-      ++pattern;
-      ++pos;				/* intentional fallthrough */
-
+      if(skip) {
+        if (*(buf+1) == '\0') {           /* but no escaping of '\0'! */
+          snprintf(glob->errormsg, sizeof(glob->errormsg),
+                   "illegal pattern at pos %d\n", (int)pos);
+          return GLOB_ERROR;
+        }
+        ++pattern;
+        ++pos;
+      }
+      /* intentional fallthrough */
     default:
-      *buf++ = *pattern++;		/* copy character to set element */
+      *buf++ = *pattern++;              /* copy character to set element */
       ++pos;
     }
   }
   /* we never reach this point */
 }
 
-static GlobCode glob_range(URLGlob *glob, char *pattern, int pos, int *amount)
+static GlobCode glob_range(URLGlob *glob, char *pattern,
+                           size_t pos, int *amount)
 {
   /* processes a range expression with the point behind the opening '['
      - char range: e.g. "a-z]", "B-Q]"
@@ -149,78 +166,101 @@ static GlobCode glob_range(URLGlob *glob, char *pattern, int pos, int *amount)
   URLPattern *pat;
   char *c;
   int wordamount=1;
-  
+  char sep;
+  char sep2;
+  int step;
+  int rc;
+
   pat = (URLPattern*)&glob->pattern[glob->size / 2];
   /* patterns 0,1,2,... correspond to size=1,3,5,... */
   ++glob->size;
 
-  if (isalpha((int)*pattern)) {		/* character range detected */
+  if (isalpha((int)*pattern)) {         /* character range detected */
+    char min_c;
+    char max_c;
+
     pat->type = UPTCharRange;
-    if (sscanf(pattern, "%c-%c]", &pat->content.CharRange.min_c,
-               &pat->content.CharRange.max_c) != 2 ||
-	pat->content.CharRange.min_c >= pat->content.CharRange.max_c ||
-	pat->content.CharRange.max_c - pat->content.CharRange.min_c > 'z' - 'a') {
-      /* the pattern is not well-formed */ 
+    rc = sscanf(pattern, "%c-%c%c%d%c", &min_c, &max_c, &sep, &step, &sep2);
+    if ((rc < 3) || (min_c >= max_c) || ((max_c - min_c) > ('z' - 'a'))) {
+      /* the pattern is not well-formed */
       snprintf(glob->errormsg, sizeof(glob->errormsg),
-               "illegal pattern or range specification after pos %d\n", pos);
+               "errpr: bad range specification after pos %d\n", pos);
       return GLOB_ERROR;
     }
-    pat->content.CharRange.ptr_c = pat->content.CharRange.min_c;
-    /* always check for a literal (may be "") between patterns */
 
-    if(GLOB_ERROR == glob_word(glob, pattern + 4, pos + 4, &wordamount))
-      wordamount=1;
+    /* check the (first) separating character */
+    if((sep != ']') && (sep != ':')) {
+      snprintf(glob->errormsg, sizeof(glob->errormsg),
+               "error: unsupported character (%c) after range at pos %d\n",
+               sep, pos);
+      return GLOB_ERROR;
+    }
 
-    *amount = (pat->content.CharRange.max_c -
-               pat->content.CharRange.min_c + 1) *
-      wordamount;
+    /* if there was a ":[num]" thing, use that as step or else use 1 */
+    pat->content.CharRange.step =
+      ((sep == ':') && (rc == 5) && (sep2 == ']'))?step:1;
 
-    return GLOB_OK;
+    pat->content.CharRange.ptr_c = pat->content.CharRange.min_c = min_c;
+    pat->content.CharRange.max_c = max_c;
   }
-
-  if (isdigit((int)*pattern)) { /* numeric range detected */
+  else if (isdigit((int)*pattern)) { /* numeric range detected */
+    int min_n;
+    int max_n;
 
     pat->type = UPTNumRange;
     pat->content.NumRange.padlength = 0;
-    if (sscanf(pattern, "%d-%d]",
-               &pat->content.NumRange.min_n,
-               &pat->content.NumRange.max_n) != 2 ||
-	pat->content.NumRange.min_n >= pat->content.NumRange.max_n) {
-      /* the pattern is not well-formed */ 
-      snprintf(glob->errormsg, sizeof(glob->errormsg), 
-               "error: illegal pattern or range specification after pos %d\n",
-               pos);
+
+    rc = sscanf(pattern, "%d-%d%c%d%c", &min_n, &max_n, &sep, &step, &sep2);
+
+    if ((rc < 2) || (min_n >= max_n)) {
+      /* the pattern is not well-formed */
+      snprintf(glob->errormsg, sizeof(glob->errormsg),
+               "error: bad range specification after pos %d\n", pos);
       return GLOB_ERROR;
     }
-    if (*pattern == '0') {		/* leading zero specified */
-      c = pattern;  
+    pat->content.NumRange.ptr_n =  pat->content.NumRange.min_n = min_n;
+    pat->content.NumRange.max_n = max_n;
+
+    /* if there was a ":[num]" thing, use that as step or else use 1 */
+    pat->content.NumRange.step =
+      ((sep == ':') && (rc == 5) && (sep2 == ']'))?step:1;
+
+    if (*pattern == '0') {              /* leading zero specified */
+      c = pattern;
       while (isdigit((int)*c++))
-	++pat->content.NumRange.padlength; /* padding length is set for all
+        ++pat->content.NumRange.padlength; /* padding length is set for all
                                               instances of this pattern */
     }
-    pat->content.NumRange.ptr_n = pat->content.NumRange.min_n;
-    c = (char*)strchr(pattern, ']'); /* continue after next ']' */
-    if(c)
-      c++;
-    else {
-      snprintf(glob->errormsg, sizeof(glob->errormsg), "missing ']'");
-      return GLOB_ERROR; /* missing ']' */
-    }
 
-    /* always check for a literal (may be "") between patterns */
-
-    if(GLOB_ERROR == glob_word(glob, c, pos + (c - pattern), &wordamount))
-      wordamount = 1;
-    
-    *amount = (pat->content.NumRange.max_n - 
-               pat->content.NumRange.min_n + 1) *
-      wordamount;
-
-    return GLOB_OK;
   }
-  snprintf(glob->errormsg, sizeof(glob->errormsg), 
-           "illegal character in range specification at pos %d\n", pos);
-  return GLOB_ERROR;
+  else {
+    snprintf(glob->errormsg, sizeof(glob->errormsg),
+             "illegal character in range specification at pos %d\n", pos);
+    return GLOB_ERROR;
+  }
+
+  c = (char*)strchr(pattern, ']'); /* continue after next ']' */
+  if(c)
+    c++;
+  else {
+    snprintf(glob->errormsg, sizeof(glob->errormsg), "missing ']'");
+    return GLOB_ERROR; /* missing ']' */
+  }
+
+  /* always check for a literal (may be "") between patterns */
+
+  if(GLOB_ERROR == glob_word(glob, c, pos + (c - pattern), &wordamount))
+    wordamount = 1;
+
+  if(pat->type == UPTCharRange)
+    *amount = (pat->content.CharRange.max_c -
+               pat->content.CharRange.min_c + 1) *
+      wordamount;
+  else
+    *amount = (pat->content.NumRange.max_n -
+               pat->content.NumRange.min_n + 1) * wordamount;
+
+  return GLOB_OK;
 }
 
 static GlobCode glob_word(URLGlob *glob, char *pattern,
@@ -228,7 +268,7 @@ static GlobCode glob_word(URLGlob *glob, char *pattern,
 {
   /* processes a literal string component of a URL
      special characters '{' and '[' branch to set/range processing functions
-   */ 
+   */
   char* buf = glob->glob_buffer;
   size_t litindex;
   GlobCode res = GLOB_OK;
@@ -247,10 +287,10 @@ static GlobCode glob_word(URLGlob *glob, char *pattern,
       /* escape character, skip '\' */
       ++pattern;
       ++pos;
-      if (*pattern == '\0')		/* but no escaping of '\0'! */
-	return GLOB_ERROR;
+      if (*pattern == '\0')             /* but no escaping of '\0'! */
+        return GLOB_ERROR;
     }
-    *buf++ = *pattern++;		/* copy character to literal */
+    *buf++ = *pattern++;                /* copy character to literal */
     ++pos;
   }
   *buf = '\0';
@@ -263,7 +303,7 @@ static GlobCode glob_word(URLGlob *glob, char *pattern,
 
   switch (*pattern) {
   case '\0':
-    break;			/* singular URL processed  */
+    break;                      /* singular URL processed  */
 
   case '{':
     /* process set pattern */
@@ -297,7 +337,7 @@ int glob_url(URLGlob** glob, char* url, int *urlnum, FILE *error)
   if(NULL == glob_buffer)
     return CURLE_OUT_OF_MEMORY;
 
-  glob_expand = (URLGlob*)malloc(sizeof(URLGlob));
+  glob_expand = (URLGlob*)calloc(sizeof(URLGlob), 1);
   if(NULL == glob_expand) {
     free(glob_buffer);
     return CURLE_OUT_OF_MEMORY;
@@ -328,20 +368,21 @@ int glob_url(URLGlob** glob, char* url, int *urlnum, FILE *error)
 
 void glob_cleanup(URLGlob* glob)
 {
-  int i, elem;
+  size_t i;
+  int elem;
 
-  for (i = glob->size - 1; i >= 0; --i) {
-    if (!(i & 1)) {	/* even indexes contain literals */
+  for (i = glob->size - 1; i < glob->size; --i) {
+    if (!(i & 1)) {     /* even indexes contain literals */
       free(glob->literal[i/2]);
     }
-    else {		/* odd indexes contain sets or ranges */
+    else {              /* odd indexes contain sets or ranges */
       if (glob->pattern[i/2].type == UPTSet) {
-	for (elem = glob->pattern[i/2].content.Set.size - 1;
+        for (elem = glob->pattern[i/2].content.Set.size - 1;
              elem >= 0;
              --elem) {
-	  free(glob->pattern[i/2].content.Set.elements[elem]);
-	}
-	free(glob->pattern[i/2].content.Set.elements);
+          free(glob->pattern[i/2].content.Set.elements[elem]);
+        }
+        free(glob->pattern[i/2].content.Set.elements);
       }
     }
   }
@@ -354,45 +395,46 @@ char *glob_next_url(URLGlob *glob)
   char *buf = glob->glob_buffer;
   URLPattern *pat;
   char *lit;
-  int i;
+  size_t i;
   size_t j;
-  int carry;
 
   if (!glob->beenhere)
     glob->beenhere = 1;
   else {
-    carry = 1;
+    bool carry = TRUE;
 
     /* implement a counter over the index ranges of all patterns,
        starting with the rightmost pattern */
-    for (i = glob->size / 2 - 1; carry && i >= 0; --i) {
-      carry = 0;
+    for (i = glob->size / 2 - 1; carry && i < glob->size; --i) {
+      carry = FALSE;
       pat = &glob->pattern[i];
       switch (pat->type) {
       case UPTSet:
-	if (++pat->content.Set.ptr_s == pat->content.Set.size) {
-	  pat->content.Set.ptr_s = 0;
-	  carry = 1;
-	}
-	break;
+        if (++pat->content.Set.ptr_s == pat->content.Set.size) {
+          pat->content.Set.ptr_s = 0;
+          carry = TRUE;
+        }
+        break;
       case UPTCharRange:
-	if (++pat->content.CharRange.ptr_c > pat->content.CharRange.max_c) {
-	  pat->content.CharRange.ptr_c = pat->content.CharRange.min_c;
-	  carry = 1;
-	}
-	break;
+        pat->content.CharRange.ptr_c += pat->content.CharRange.step;
+        if (pat->content.CharRange.ptr_c > pat->content.CharRange.max_c) {
+          pat->content.CharRange.ptr_c = pat->content.CharRange.min_c;
+          carry = TRUE;
+        }
+        break;
       case UPTNumRange:
-	if (++pat->content.NumRange.ptr_n > pat->content.NumRange.max_n) {
-	  pat->content.NumRange.ptr_n = pat->content.NumRange.min_n;
-	  carry = 1;
-	}
-	break;
+        pat->content.NumRange.ptr_n += pat->content.NumRange.step;
+        if (pat->content.NumRange.ptr_n > pat->content.NumRange.max_n) {
+          pat->content.NumRange.ptr_n = pat->content.NumRange.min_n;
+          carry = TRUE;
+        }
+        break;
       default:
-	printf("internal error: invalid pattern type (%d)\n", (int)pat->type);
-	exit (CURLE_FAILED_INIT);
+        printf("internal error: invalid pattern type (%d)\n", (int)pat->type);
+        exit (CURLE_FAILED_INIT);
       }
     }
-    if (carry)		/* first pattern ptr has run into overflow, done! */
+    if (carry)          /* first pattern ptr has run into overflow, done! */
       return NULL;
   }
 
@@ -402,24 +444,24 @@ char *glob_next_url(URLGlob *glob)
       strcpy(buf, lit);
       buf += strlen(lit);
     }
-    else {				/* the rest (i odd) are patterns */
+    else {                              /* the rest (i odd) are patterns */
       pat = &glob->pattern[j/2];
       switch(pat->type) {
       case UPTSet:
-	strcpy(buf, pat->content.Set.elements[pat->content.Set.ptr_s]);
-	buf += strlen(pat->content.Set.elements[pat->content.Set.ptr_s]);
-	break;
+        strcpy(buf, pat->content.Set.elements[pat->content.Set.ptr_s]);
+        buf += strlen(pat->content.Set.elements[pat->content.Set.ptr_s]);
+        break;
       case UPTCharRange:
-	*buf++ = pat->content.CharRange.ptr_c;
-	break;
+        *buf++ = pat->content.CharRange.ptr_c;
+        break;
       case UPTNumRange:
-	sprintf(buf, "%0*d",
-                pat->content.NumRange.padlength, pat->content.NumRange.ptr_n); 
+        sprintf(buf, "%0*d",
+                pat->content.NumRange.padlength, pat->content.NumRange.ptr_n);
         buf += strlen(buf); /* make no sprint() return code assumptions */
-	break;
+        break;
       default:
-	printf("internal error: invalid pattern type (%d)\n", (int)pat->type);
-	exit (CURLE_FAILED_INIT);
+        printf("internal error: invalid pattern type (%d)\n", (int)pat->type);
+        exit (CURLE_FAILED_INIT);
       }
     }
   }
@@ -447,10 +489,9 @@ char *glob_match_url(char *filename, URLGlob *glob)
 
   while (*filename) {
     if (*filename == '#' && isdigit((int)filename[1])) {
-      /* only '#1' ... '#9' allowed */
       unsigned long i;
+      char *ptr = filename;
       unsigned long num = strtoul(&filename[1], &filename, 10);
-
       i = num-1;
 
       if (num && (i <= glob->size / 2)) {
@@ -480,6 +521,12 @@ char *glob_match_url(char *filename, URLGlob *glob)
           return NULL;
         }
       }
+      else {
+        /* #[num] out of range, use the #[num] in the output */
+        filename = ptr;
+        appendthis=filename++;
+        appendlen=1;
+      }
     }
     else {
       appendthis=filename++;
